package jp.gr.java_conf.coskx.ocv_gl2d;

import android.annotation.TargetApi;
import android.app.Activity;
import android.content.Context;
import android.graphics.ImageFormat;
import android.hardware.camera2.CameraAccessException;
import android.hardware.camera2.CameraCaptureSession;
import android.hardware.camera2.CameraCharacteristics;
import android.hardware.camera2.CameraDevice;
import android.hardware.camera2.CameraManager;
import android.hardware.camera2.CaptureRequest;
import android.hardware.camera2.params.OutputConfiguration;
import android.hardware.camera2.params.SessionConfiguration;
import android.hardware.camera2.params.StreamConfigurationMap;
import android.media.Image;
import android.media.ImageReader;
import android.os.Build;
import android.os.Handler;
import android.os.HandlerThread;
import android.os.Looper;
import android.util.AttributeSet;
import android.util.Log;
import android.view.Display;
import android.view.Surface;
import android.view.ViewGroup.LayoutParams;
import android.widget.TextView;

import androidx.annotation.RequiresApi;

import org.opencv.core.CvType;
import org.opencv.core.Mat;
import org.opencv.core.Size;
import org.opencv.imgproc.Imgproc;

import java.nio.ByteBuffer;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.concurrent.Executors;

/**
 * This class is an implementation of the Bridge View between OpenCV and Java Camera.
 * This class relays on the functionality available in base class and only implements
 * required functions:
 * connectCamera - opens Java camera and sets the PreviewCallback to be delivered.
 * disconnectCamera - closes the camera and stops preview.
 * When frame is delivered via callback from Camera - it processed via OpenCV to be
 * converted to RGBA32 and then passed to the external callback for modifications if required.
 */

@TargetApi(21)
public class JavaCamera2ViewGL extends CameraBridgeViewBaseGL {

    private static final String TAG = "JavaCamera2ViewGL";

    private ImageReader mImageReader;
    private int mPreviewFormat = ImageFormat.YUV_420_888;

    private CameraDevice mCameraDevice;
    private CameraCaptureSession mCaptureSession;
    private CaptureRequest.Builder mPreviewRequestBuilder;
    private String mCameraID;
    private android.util.Size mPreviewSize = new android.util.Size(-1, -1);

    private HandlerThread mBackgroundThread;
    private Handler mBackgroundHandler;
    private Handler mHandler;

    private TextView info2;

    public JavaCamera2ViewGL(Context context, int cameraId) {
        super(context, cameraId);
    }

    public JavaCamera2ViewGL(Context context, AttributeSet attrs) {
        super(context, attrs);
    }

    public void setFPSTextView(TextView tv){
        info2 = tv;
    }

    private void startBackgroundThread() {
        Log.i(TAG, "startBackgroundThread()");
        stopBackgroundThread();
        mBackgroundThread = new HandlerThread("OpenCVCameraBackground");
        mBackgroundThread.start();
        mBackgroundHandler = new Handler(mBackgroundThread.getLooper());
    }

    private void stopBackgroundThread() {
        Log.i(TAG, "stopBackgroundThread()");
        if (mBackgroundThread == null)
            return;
        mBackgroundThread.quitSafely();
        try {
            mBackgroundThread.join();
            mBackgroundThread = null;
            mBackgroundHandler = null;
        } catch (InterruptedException e) {
            Log.e(TAG, "stopBackgroundThread", e);
        }
    }

    protected boolean initializeCamera() {
        Log.i(TAG, "initializeCamera()");
        CameraManager manager = (CameraManager) getContext().getSystemService(Context.CAMERA_SERVICE);
        try {
            String camList[] = manager.getCameraIdList();
            if (camList.length == 0) {
                Log.e(TAG, "Error: camera isn't detected.");
                return false;
            }
            if (mCameraIndex == CameraBridgeViewBaseGL.CAMERA_ID_ANY) {
                mCameraID = camList[0];
            } else {
                for (String cameraID : camList) {
                    CameraCharacteristics characteristics = manager.getCameraCharacteristics(cameraID);
                    if ((mCameraIndex == CameraBridgeViewBaseGL.CAMERA_ID_BACK &&
                            characteristics.get(CameraCharacteristics.LENS_FACING) == CameraCharacteristics.LENS_FACING_BACK) ||
                            (mCameraIndex == CameraBridgeViewBaseGL.CAMERA_ID_FRONT &&
                                    characteristics.get(CameraCharacteristics.LENS_FACING) == CameraCharacteristics.LENS_FACING_FRONT)
                    ) {
                        mCameraID = cameraID;
                        break;
                    }
                }
            }
            if (mCameraID != null) {
                Log.i(TAG, "initializeCamera() P1 Opening camera: " + mCameraID);
                manager.openCamera(mCameraID, mStateCallback, mBackgroundHandler);
            } else { // make JavaCamera2ViewGL behaves in the same way as JavaCameraView
                Log.i(TAG, "initializeCamera() Trying to open camera with the value (" + mCameraIndex + ")");
                if (mCameraIndex < camList.length) {
                    mCameraID = camList[mCameraIndex];
                    Log.i(TAG, "initializeCamera() P2 Opening camera: " + mCameraID);
                    manager.openCamera(mCameraID, mStateCallback, mBackgroundHandler);
                } else {
                    // CAMERA_DISCONNECTED is used when the camera id is no longer valid
                    throw new CameraAccessException(CameraAccessException.CAMERA_DISCONNECTED);
                }
            }
            return true;
        } catch (CameraAccessException e) {
            Log.e(TAG, "OpenCamera - Camera Access Exception", e);
        } catch (IllegalArgumentException e) {
            Log.e(TAG, "OpenCamera - Illegal Argument Exception", e);
        } catch (SecurityException e) {
            Log.e(TAG, "OpenCamera - Security Exception", e);
        }
        return false;
    }

    private final CameraDevice.StateCallback mStateCallback = new CameraDevice.StateCallback() {

        @Override
        public void onOpened(CameraDevice cameraDevice) {
            Log.i(TAG, "call back onOpened()  calling createCameraPreviewSession()");
            mCameraDevice = cameraDevice;
            createCameraPreviewSession();
            mHandler = new Handler(Looper.getMainLooper());
        }

        @Override
        public void onDisconnected(CameraDevice cameraDevice) {
            Log.i(TAG, "call back onDisconnected()");
            cameraDevice.close();
            mCameraDevice = null;
            mHandler.removeCallbacksAndMessages(null);
        }

        @Override
        public void onError(CameraDevice cameraDevice, int error) {
            cameraDevice.close();
            mCameraDevice = null;
        }

    };

    private boolean PreviewSessioniscreating = false;

    private void createCameraPreviewSession() {
        final int w = mPreviewSize.getWidth(), h = mPreviewSize.getHeight();
        Log.i(TAG, "createCameraPreviewSession(" + w + "x" + h + ")");
        if (w < 0 || h < 0)
            return;
        if (null == mCameraDevice) {
            Log.e(TAG, "createCameraPreviewSession: camera isn't opened");
            return;
        }
        if (null != mCaptureSession) {
            Log.e(TAG, "createCameraPreviewSession: mCaptureSession is already started");
            return;
        }

        Log.i(TAG, "createCameraPreviewSession: set ImageReader w x h = " + w +" x " + h);
        mImageReader = ImageReader.newInstance(w, h, mPreviewFormat, 2);
        mImageReader.setOnImageAvailableListener(new ImageReader.OnImageAvailableListener() {
            @Override
            public void onImageAvailable(ImageReader reader) {
                Image image = reader.acquireLatestImage();
                if (image == null)
                    return;

                // sanity checks - 3 planes
                Image.Plane[] planes = image.getPlanes();
                assert (planes.length == 3);
                assert (image.getFormat() == mPreviewFormat);

                JavaCamera2Frame tempFrame = new JavaCamera2Frame(image);
                deliverAndDrawFrame(tempFrame);
                tempFrame.release();
                image.close();
            }
        }, mBackgroundHandler);
        Surface surface = mImageReader.getSurface();

        try {
            mPreviewRequestBuilder = mCameraDevice.createCaptureRequest(CameraDevice.TEMPLATE_PREVIEW);
            mPreviewRequestBuilder.addTarget(surface);
        } catch (CameraAccessException e) {
            Log.e(TAG, "createCameraPreviewSession", e);
        }
        CameraCaptureSession.StateCallback callback= setCameraCaptureSessionStateCB(surface);

        if (android.os.Build.VERSION.SDK_INT >= android.os.Build.VERSION_CODES.P) {
            setCameraDevicecreateCaptureSessionPandLater(surface,callback);
        } else {
            setCameraDevicecreateCaptureSessionBeforeP(surface,callback);
        }
    }

    private CameraCaptureSession.StateCallback setCameraCaptureSessionStateCB(Surface surface) {
        return new CameraCaptureSession.StateCallback() {
            @Override
            public void onConfigured(CameraCaptureSession cameraCaptureSession) {
                Log.i(TAG, "createCaptureSession::onConfigured");
                if (null == mCameraDevice) {
                    return; // camera is already closed
                }
                mCaptureSession = cameraCaptureSession;
                try {
                    mPreviewRequestBuilder.set(CaptureRequest.CONTROL_AF_MODE,
                            CaptureRequest.CONTROL_AF_MODE_CONTINUOUS_PICTURE);
                    mPreviewRequestBuilder.set(CaptureRequest.CONTROL_AE_MODE,
                            CaptureRequest.CONTROL_AE_MODE_ON_AUTO_FLASH);

                    mCaptureSession.setRepeatingRequest(mPreviewRequestBuilder.build(), null, mBackgroundHandler);
                    Log.i(TAG, "CameraPreviewSession has been started");
                } catch (Exception e) {
                    Log.e(TAG, "createCaptureSession failed", e);
                }
            }

            @Override
            public void onConfigureFailed(CameraCaptureSession cameraCaptureSession) {
                Log.e(TAG, "createCameraPreviewSession failed");
            }
        };
    }

    @RequiresApi(api = Build.VERSION_CODES.P)
    private void setCameraDevicecreateCaptureSessionPandLater(Surface surface, CameraCaptureSession.StateCallback callback) {
        try {
            mCameraDevice.createCaptureSession(
                    new SessionConfiguration(
                            SessionConfiguration.SESSION_REGULAR,
                            Collections.singletonList(new OutputConfiguration(surface)),
                            Executors.newCachedThreadPool(),
                            callback
                    ));
        } catch (CameraAccessException e) {
            e.printStackTrace();
        }
    }

    @SuppressWarnings("deprecation")
    private void setCameraDevicecreateCaptureSessionBeforeP(Surface surface, CameraCaptureSession.StateCallback callback) {
        try {
            mCameraDevice.createCaptureSession(Arrays.asList(surface), callback, null);
        } catch (CameraAccessException e) {
            e.printStackTrace();
        }
    }

    @Override
    protected void disconnectCamera() {
        Log.i(TAG, "disconnectCamera() close camera");
        try {
            CameraDevice c = mCameraDevice;
            mCameraDevice = null;
            if (null != mCaptureSession) {
                mCaptureSession.close();
                mCaptureSession = null;
            }
            if (null != c) {
                c.close();
            }
        } finally {
            stopBackgroundThread();
            if (null != mImageReader) {
                mImageReader.close();
                mImageReader = null;
            }
        }
        PreviewSessioniscreating = false;
        Log.i(TAG, "disconnectCamera() camera closed!");
    }

    public static class JavaCameraSizeAccessor implements ListItemAccessor {
        @Override
        public int getWidth(Object obj) {
            android.util.Size size = (android.util.Size)obj;
            return size.getWidth();
        }

        @Override
        public int getHeight(Object obj) {
            android.util.Size size = (android.util.Size)obj;
            return size.getHeight();
        }
    }

    boolean calcPreviewSize(final int width, final int height) {
        Log.i(TAG, "calcPreviewSize(): " + width + "x" + height);
        if (mCameraID == null) {
            Log.e(TAG, "Camera isn't initialized!");
            return false;
        }
        if (isPortrait) {
            longerSide = height;
            shorterSide = width;
        } else {
            longerSide = width;
            shorterSide = height;
        }

        CameraManager manager = (CameraManager) getContext().getSystemService(Context.CAMERA_SERVICE);
        try {
            CameraCharacteristics characteristics = manager.getCameraCharacteristics(mCameraID);
            StreamConfigurationMap map = characteristics.get(CameraCharacteristics.SCALER_STREAM_CONFIGURATION_MAP);
            android.util.Size[] sizes = map.getOutputSizes(ImageReader.class);
            List<android.util.Size> sizes_list = Arrays.asList(sizes);
            //Size frameSize = calculateCameraFrameSize(sizes_list, new JavaCameraSizeAccessor(), width, height);
            Size frameSize = calculateCameraFrameSize(sizes_list, new JavaCameraSizeAccessor(), longerSide, shorterSide);
            Log.i(TAG, "calcPreviewSize() Selected preview size to " + Integer.valueOf((int)frameSize.width) + "x" + Integer.valueOf((int)frameSize.height));
            assert(!(frameSize.width == 0 || frameSize.height == 0));
            if (mPreviewSize.getWidth() == frameSize.width && mPreviewSize.getHeight() == frameSize.height) {
                Log.i(TAG, "calcPreviewSize(): return false");
                return false;
            } else {
                mPreviewSize = new android.util.Size((int)frameSize.width, (int)frameSize.height);
                Log.i(TAG, "calcPreviewSize(): return true");
                return true;
            }
        } catch (CameraAccessException e) {
            Log.e(TAG, "calcPreviewSize - Camera Access Exception", e);
        } catch (IllegalArgumentException e) {
            Log.e(TAG, "calcPreviewSize - Illegal Argument Exception", e);
        } catch (SecurityException e) {
            Log.e(TAG, "calcPreviewSize - Security Exception", e);
        }
        return false;
    }

    @Override
    protected boolean connectCamera(int width, int height) {
        Log.i(TAG, "connectCamera() setCameraPreviewSize(" + width + "x" + height + ")");
        if ( width < height) isPortrait = true;
        else  isPortrait = false;
        startBackgroundThread();
        initializeCamera();
        try {
            boolean needReconfig = calcPreviewSize(width, height);
            mFrameWidth = mPreviewSize.getWidth();
            mFrameHeight = mPreviewSize.getHeight();
            Log.i(TAG,"connectCamera() mFrameWidth mFrameHeight = " + mFrameWidth +" x " + mFrameHeight); //縦のとき1080 x 1440

            if ((getLayoutParams().width == LayoutParams.MATCH_PARENT) && (getLayoutParams().height == LayoutParams.MATCH_PARENT)) {
                if (isPortrait) {
                    mScaleW = (float) height / mFrameWidth;
                    mScaleH = (float) width / mFrameHeight;
                } else {
                    mScaleW = (float) width / mFrameWidth;
                    mScaleH = (float) height / mFrameHeight;
                }
                Log.i(TAG,"connectCamera() mScaleW mScaleH = " + mScaleW +" x " + mScaleH); //縦のとき1080 x 1440
                mScale = Math.min(mScaleW, mScaleH);
            } else
                mScale = 0;
            Log.i(TAG,"connectCamera()  mscale = "+mScale);

            AllocateCache();

            if (needReconfig) {
                if (null != mCaptureSession) {
                    Log.d(TAG, "connectCamera() closing existing previewSession");
                    mCaptureSession.close();
                    mCaptureSession = null;
                }
                createCameraPreviewSession();
                Log.i(TAG,"connectCamera() just return from createCameraPreviewSession");
            }
        } catch (RuntimeException e) {
            throw new RuntimeException("Interrupted while setCameraPreviewSize.", e);
        }
        rotationindex = getRotationIndex();
        Log.i(TAG,"connectCamera() rotationindex = " + rotationindex);
        mGLRenderer.makeSettings(rotationindex, mScaleW, mScaleH);
        return true;
    }

    private void setRotationIndex() {
        rotationindex = getRotationIndex();  //rotationindex = 0; //0:なにもしない　1:CW90回す　2:180回す　3:CCW90回す
        Log.i(TAG,"getRotationIndex() rotationindex = " + rotationindex);
        if(rotationindex != rotationindexPrev) {
            mGLRenderer.makeSettings(rotationindex, mScaleW, mScaleH);
            rotationindexPrev = rotationindex;
        }
    }

    private class JavaCamera2Frame implements CvCameraViewFrame {
        @Override
        public Mat gray() {
            Image.Plane[] planes = mImage.getPlanes();
            int w = mImage.getWidth();
            int h = mImage.getHeight();
            assert(planes[0].getPixelStride() == 1);
            ByteBuffer y_plane = planes[0].getBuffer();
            int y_plane_step = planes[0].getRowStride();
            mGray = new Mat(h, w, CvType.CV_8UC1, y_plane, y_plane_step);
            return mGray;
        }

        @Override
        public Mat rgba() {
            Image.Plane[] planes = mImage.getPlanes();
            int w = mImage.getWidth();
            int h = mImage.getHeight();
            int chromaPixelStride = planes[1].getPixelStride();


            if (chromaPixelStride == 2) { // Chroma channels are interleaved
                assert(planes[0].getPixelStride() == 1);
                assert(planes[2].getPixelStride() == 2);
                ByteBuffer y_plane = planes[0].getBuffer();
                int y_plane_step = planes[0].getRowStride();
                ByteBuffer uv_plane1 = planes[1].getBuffer();
                int uv_plane1_step = planes[1].getRowStride();
                ByteBuffer uv_plane2 = planes[2].getBuffer();
                int uv_plane2_step = planes[2].getRowStride();
                Mat y_mat = new Mat(h, w, CvType.CV_8UC1, y_plane, y_plane_step);
                Mat uv_mat1 = new Mat(h / 2, w / 2, CvType.CV_8UC2, uv_plane1, uv_plane1_step);
                Mat uv_mat2 = new Mat(h / 2, w / 2, CvType.CV_8UC2, uv_plane2, uv_plane2_step);
                long addr_diff = uv_mat2.dataAddr() - uv_mat1.dataAddr();
                if (addr_diff > 0) {
                    assert(addr_diff == 1);
                    Imgproc.cvtColorTwoPlane(y_mat, uv_mat1, mRgba, Imgproc.COLOR_YUV2RGBA_NV12);
                } else {
                    assert(addr_diff == -1);
                    Imgproc.cvtColorTwoPlane(y_mat, uv_mat2, mRgba, Imgproc.COLOR_YUV2RGBA_NV21);
                }
                return mRgba;
            } else { // Chroma channels are not interleaved
                byte[] yuv_bytes = new byte[w*(h+h/2)];
                ByteBuffer y_plane = planes[0].getBuffer();
                ByteBuffer u_plane = planes[1].getBuffer();
                ByteBuffer v_plane = planes[2].getBuffer();

                int yuv_bytes_offset = 0;

                int y_plane_step = planes[0].getRowStride();
                if (y_plane_step == w) {
                    y_plane.get(yuv_bytes, 0, w*h);
                    yuv_bytes_offset = w*h;
                } else {
                    int padding = y_plane_step - w;
                    for (int i = 0; i < h; i++){
                        y_plane.get(yuv_bytes, yuv_bytes_offset, w);
                        yuv_bytes_offset += w;
                        if (i < h - 1) {
                            y_plane.position(y_plane.position() + padding);
                        }
                    }
                    assert(yuv_bytes_offset == w * h);
                }

                int chromaRowStride = planes[1].getRowStride();
                int chromaRowPadding = chromaRowStride - w/2;

                if (chromaRowPadding == 0){
                    // When the row stride of the chroma channels equals their width, we can copy
                    // the entire channels in one go
                    u_plane.get(yuv_bytes, yuv_bytes_offset, w*h/4);
                    yuv_bytes_offset += w*h/4;
                    v_plane.get(yuv_bytes, yuv_bytes_offset, w*h/4);
                } else {
                    // When not equal, we need to copy the channels row by row
                    for (int i = 0; i < h/2; i++){
                        u_plane.get(yuv_bytes, yuv_bytes_offset, w/2);
                        yuv_bytes_offset += w/2;
                        if (i < h/2-1){
                            u_plane.position(u_plane.position() + chromaRowPadding);
                        }
                    }
                    for (int i = 0; i < h/2; i++){
                        v_plane.get(yuv_bytes, yuv_bytes_offset, w/2);
                        yuv_bytes_offset += w/2;
                        if (i < h/2-1){
                            v_plane.position(v_plane.position() + chromaRowPadding);
                        }
                    }
                }

                Mat yuv_mat = new Mat(h+h/2, w, CvType.CV_8UC1);
                yuv_mat.put(0, 0, yuv_bytes);
                Imgproc.cvtColor(yuv_mat, mRgba, Imgproc.COLOR_YUV2RGBA_I420, 4);
                return mRgba;
            }
        }


        public JavaCamera2Frame(Image image) {
            super();
            mImage = image;
            mRgba = new Mat();
            mGray = new Mat();
        }

        public void release() {
            mRgba.release();
            mGray.release();
        }

        private Image mImage;
        private Mat mRgba;
        private Mat mGray;
    };

    Activity mActivity;
    public void setActivity(Activity act) {
        mActivity = act;
    }


    public int getRotationIndex() {
        int camerarotation=0; //カメラの向き　90度で正規化
        int facerotation=0;   //backcamera:0 frontcamera:2
        int devicerotation=0;  //デバイスの向き　90度で正規化
        CameraManager manager = (CameraManager) getContext().getSystemService(Context.CAMERA_SERVICE);
        try {
            CameraCharacteristics characteristics = manager.getCameraCharacteristics(mCameraID);
            if (characteristics.get(CameraCharacteristics.LENS_FACING) == CameraCharacteristics.LENS_FACING_BACK) facerotation = 0;
            else facerotation = 2;
            camerarotation = characteristics.get(CameraCharacteristics.SENSOR_ORIENTATION)/90;
        } catch (CameraAccessException e) {
            Log.e(TAG, "getRotationIndex - Camera Access Exception", e);
        } catch (IllegalArgumentException e) {
            Log.e(TAG, "getRotationIndex - Illegal Argument Exception", e);
        } catch (SecurityException e) {
            Log.e(TAG, "getRotationIndex - Security Exception", e);
        }
        devicerotation = getMyDisplay().getRotation();
        Log.i(TAG,"getRotationIndex() facerotation, camerarotation, devicerotation = " + facerotation + ", " + camerarotation + ", " + devicerotation);
        // face  camera   device     rotindex
        //P10lite
        //   0      1       0        1  (時計回り90度)
        //   0      1       1        0
        //   0      1       2        NU
        //   0      1       3        2  (180度)
        //Nexus5x
        //   0      3       0        3 　(反時計回り90度)
        //   0      3       1        2  (180度)
        //   0      3       2        NU
        //   0      3       3        0
        //P10lite
        //   2      3       0        3 　(反時計回り90度)
        //   2      3       1        0
        //   2      3       2        NU
        //   2      3       3        2  (180度)
        //Nexus5x
        //   2      3       0        3　　(反時計回り90度)
        //   2      3       1        0
        //   2      3       2        NU
        //   2      3       3        2  (180度)
        //仮想機種
        //   2      1       0        1　　(時計回り90度)
        //   2      1       1        2  (180度)
        //   2      1       2        NU
        //   2      1       3        0
        int rotationindex=0;
        if (isPortrait) rotationindex = camerarotation;
        else rotationindex = (facerotation + camerarotation - devicerotation + 4) % 4;
        return rotationindex;
    }

    private long previoustime = 0;
    private int loopcount=0;
    private String fpsstring="";
    protected void showFPS(){
        long currenttime = 0;
        if (previoustime == 0) {
            previoustime = System.currentTimeMillis();
            loopcount = 0;
        } else if (loopcount == 50) {
            currenttime = System.currentTimeMillis();
            double fps = 50./((double)(currenttime - previoustime)/1000.);
            Log.i(TAG, "FPS = " + fps);
            fpsstring = String.format(" %4.1fFPS @%dx%d ", fps, mFrameWidth, mFrameHeight);
            previoustime = currenttime;
            loopcount = 0;
            mHandler.post(new Runnable() {
                @Override
                public void run() {
                    info2.setText(fpsstring);
                    return;
                }
            });
        }
        loopcount++;
    }

    @SuppressWarnings("deprecation")
    private Display getMyDisplay() {
        Display mDisplay;
        if (android.os.Build.VERSION.SDK_INT >= android.os.Build.VERSION_CODES.R) {
            mDisplay =this.getDisplay();
        } else {
            mDisplay = mActivity.getWindowManager().getDefaultDisplay();
        }
        return mDisplay;
    }

    public void enableViewWithclear() {
        mPreviewSize = new android.util.Size(-1, -1);
        enableView();
    }
}
